package org.processmining.divideandconquer.algorithms;

import org.deckfour.xes.extension.std.XConceptExtension;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.acceptingpetrinet.models.AcceptingPetriNet;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.decomposedminer.parameters.DecomposedDiscoveryParameters;
import org.processmining.decomposedminer.plugins.DecomposedDiscoveryPlugin;
import org.processmining.decomposedreplayer.parameters.DecomposedReplayParameters;
import org.processmining.decomposedreplayer.plugins.DecomposedReplayPlugin;
import org.processmining.divideandconquer.extensions.DrFurbyExtension;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.log.utils.XUtils;
import org.processmining.logdecomposer.classifiers.StartEndClassifier;
import org.processmining.plugins.petrinet.replayresult.PNRepResult;
import org.processmining.plugins.replayer.replayresult.SyncReplayResult;

public class DrFurbyAlgorithm {

	private static DecomposedDiscoveryPlugin discoveryPlugin = new DecomposedDiscoveryPlugin();
	private static DecomposedReplayPlugin replayPlugin = new DecomposedReplayPlugin();

	public XLog apply(final PluginContext context, XLog trainingLog, XLog testLog) {

		long time = -System.currentTimeMillis();
		context.getProgress().setIndeterminate(false);
		context.getProgress().setMinimum(0);
		context.getProgress().setMaximum(5);
		context.getProgress().setValue(0);
		
		XFactory factory = XFactoryRegistry.instance().currentDefault();
		XLog classifiedLog = (XLog) testLog.clone();
		DrFurbyExtension extension = DrFurbyExtension.instance();

		double costsIM100[] = test(context, trainingLog, classifiedLog, "Inductive Miner (Perfect Fitness)", 100);
		double costsHIM0[] = test(context, trainingLog, classifiedLog, "Hybrid ILP MIner (Default)", 0);
		
		double costs[] = new double[classifiedLog.size()];
		String log[] = new String[classifiedLog.size()];
		for (int i = 0; i < classifiedLog.size(); i++) {
			costs[i] = costsIM100[i] + costsHIM0[i];
			log[i] = "(" + costsIM100[i] + ", " + costsHIM0[i] + ")";
		}
		
		context.log("Wrapping up");
		/*
		 * Output (non-)fitting traces, and number of fitting traces.
		 */
		long fit = 0;
		for (int i = 0; i < classifiedLog.size(); i++) {
			if (costs[i] == 0.0) {
				fit++;
			}
		}
		int i = 0;
		for (XTrace trace : classifiedLog) {
			XConceptExtension.instance().assignName(trace, XConceptExtension.instance().extractName(trace) + (costs[i] == 0.0 ? " +" : ""));
			extension.assignClass(trace, costs[i] == 0.0 ? "positive" : "negative");
			extension.assignTotal(trace, costs[i]);
			extension.assignIM100(trace, costsIM100[i]);
			extension.assignHIM0(trace, costsHIM0[i]);
			String label = XConceptExtension.instance().extractName(trace);
			if (label == null) {
				label = "" + (i + 1);
			}
			System.out.println("Trace " + label + ": " + (costs[i] == 0.0 ? "fit" : "no fit, costs " + costs[i]) + " " + log[i]);
			i++;
		}
		System.out.println("Fitting traces: " + fit);
		
		System.out.println("[DrFurbyAlgorithm] @74");
		XUtils.renameLogWithProMLabel(context, trainingLog);
		System.out.println("[DrFurbyAlgorithm] @76 " + XConceptExtension.instance().extractName(trainingLog));
		XUtils.renameLogWithProMLabel(context, testLog);
		System.out.println("[DrFurbyAlgorithm] @78 " + XConceptExtension.instance().extractName(testLog));
		extension.assignName(classifiedLog, XConceptExtension.instance().extractName(trainingLog));
		System.out.println("[DrFurbyAlgorithm] @80");
		XConceptExtension.instance().assignName(classifiedLog, XConceptExtension.instance().extractName(testLog) + " (DrFurby Classified)");
		System.out.println("[DrFurbyAlgorithm] @82 " + XConceptExtension.instance().extractName(classifiedLog));
		extension.assignPos(classifiedLog, fit);
		System.out.println("[DrFurbyAlgorithm] @84");
		extension.assignNeg(classifiedLog, classifiedLog.size() - fit);
		System.out.println("[DrFurbyAlgorithm] @86");
//		context.getProgress().inc();
//		System.out.println("[DrFurbyAlgorithm] @88");
		time += System.currentTimeMillis();
		System.out.println("[DrFurbyAlgorithm] @90");
		extension.assignMillis(classifiedLog, time);
		System.out.println("[DrFurbyAlgorithm] @92");
		return classifiedLog;
	}

	private double[] test(final PluginContext context, final XLog train, final XLog test, final String discoverer, final int decomposition) {
		/*
		 * Discover a net from the training log.
		 */
		context.log("Train using " + discoverer + " and " + decomposition);
		AcceptingPetriNet discoveredNet = discover(context, train, discoverer, decomposition);
		/*
		 * Add the model as favorite to the workspace.
		 */
		context.getProvidedObjectManager().createProvidedObject(
				"Model using " + discoverer + " and " + decomposition,
				discoveredNet, AcceptingPetriNet.class, context);
		if (context instanceof UIPluginContext) {
			((UIPluginContext) context).getGlobalContext().getResourceManager()
					.getResourceForInstance(discoveredNet).setFavorite(true);
		}
		context.getProgress().inc();
		/*
		 * Replay the testlog on the discovered net
		 */
		context.log("Test");
		PNRepResult repResult = replay(context, test, discoveredNet);
		double costs[] = new double[test.size()];
		for (SyncReplayResult traceResult : repResult) {
			for (int i : traceResult.getTraceIndex()) {
				double c = traceResult.getInfo().get(PNRepResult.RAWFITNESSCOST);
				costs[i] += c;
			}
		}
		context.getProgress().inc();
		return costs;
	}
	
	private AcceptingPetriNet discover(final PluginContext context, final XLog log, final String discoverer, final int decomposition) {
		System.out.println("Discovery " + discoverer + " " + decomposition);
		DecomposedDiscoveryParameters discoveryParameters = new DecomposedDiscoveryParameters(log);
		if (decomposition == 0) {
			discoveryParameters.setConfiguration("Do not decompose");
		} else {
			discoveryParameters.setConfiguration("Decompose");
			discoveryParameters.setPercentage(decomposition);
		}
		discoveryParameters.setMiner(discoverer);
		discoveryParameters.setTryConnections(false);
		discoveryParameters.setClassifier(new StartEndClassifier(discoveryParameters.getClassifier()));
		long time = -System.currentTimeMillis();
		AcceptingPetriNet result = discoveryPlugin.run(context, log, discoveryParameters);
		time += System.currentTimeMillis();
		System.out.println("Discovery took " + time + " milliseconds.");
		return result;
	}

	private PNRepResult replay(final PluginContext context, final XLog log, final AcceptingPetriNet apn) {
		System.out.println("[Replay]");
		DecomposedReplayParameters replayParameters = new DecomposedReplayParameters(log, apn);
		replayParameters.setConfiguration("Decompose");
		long time = -System.currentTimeMillis();
		PNRepResult result = replayPlugin.apply(context, log, apn, replayParameters);
		time += System.currentTimeMillis();
		System.out.println("Replay took " + time + " milliseconds.");
		return result;
	}

}
